use std::io::Write;
use std::path::PathBuf;
use std::time::{Duration, SystemTime};

use chrono::{offset::Utc, DateTime};
use clap::Parser;
use serde::{Deserialize, Serialize};

const SECONDS_IN_MINUTE: u64 = 60;

#[derive(clap::ValueEnum, Clone, Debug)]
enum ReportFormat {
    Markdown,
    #[allow(clippy::upper_case_acronyms)]
    CSV,
}

#[derive(clap::ValueEnum, Clone, Debug)]
enum ReportTimescale {
    Hours,
    Minutes,
}

impl std::fmt::Display for ReportTimescale {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ReportTimescale::Hours => write!(f, "h"),
            ReportTimescale::Minutes => write!(f, "m"),
        }
    }
}

impl ReportTimescale {
    fn header(&self) -> &'static str {
        match self {
            ReportTimescale::Hours => "Hours",
            ReportTimescale::Minutes => "Minutes",
        }
    }

    fn divisor(&self) -> f64 {
        match self {
            ReportTimescale::Hours => 60f64 * 60f64,
            ReportTimescale::Minutes => 60f64,
        }
    }
}

#[derive(Parser)]
enum PunchIn {
    Start {
        /// The file to store time-tracking information.
        file: PathBuf,

        /// Round up to this many minutes.
        #[arg(long, default_value = "30")]
        round_up_to: u64,
    },

    Report {
        /// The format to generate the report in.
        #[arg(long, value_enum, default_value_t = ReportFormat::Markdown)]
        format: ReportFormat,

        /// The timescale to represent tracked time as.
        #[arg(long, value_enum, default_value_t = ReportTimescale::Hours)]
        timescale: ReportTimescale,

        /// The file to store time-tracking information.
        file: PathBuf,
    },
}

#[derive(Serialize, Deserialize, Clone, Debug)]
struct TimeTracking {
    start: SystemTime,
    end: Duration,
    project: String,
}

fn main() {
    let args = PunchIn::parse();

    match args {
        PunchIn::Start {
            file: path,
            round_up_to,
        } => {
            let tt = if std::fs::exists(&path).is_ok_and(|v| v) {
                let file = std::fs::File::open(&path).unwrap();
                let mut tt: Vec<TimeTracking> = serde_json::from_reader(&file).unwrap();
                start(&mut tt, round_up_to);
                let mut new_path = path.clone();
                let mut new_filename = new_path
                    .components()
                    .last()
                    .unwrap()
                    .as_os_str()
                    .to_os_string();
                new_path.pop();
                new_filename.push(".bak");
                new_path.push(new_filename);
                std::fs::rename(&path, new_path).unwrap();
                tt
            } else {
                let mut tt = vec![];
                start(&mut tt, round_up_to);
                tt
            };
            println!("Writing to file: {tt:?}");
            std::fs::write(&path, serde_json::to_string(&tt).unwrap()).unwrap();
        }
        PunchIn::Report {
            format: ReportFormat::Markdown,
            timescale,
            file: path,
        } => {
            let header = timescale.header();
            println!("| Date       | {header} | Project |");
            println!(
                "|------------|-{}-|---------|",
                String::from("-").repeat(header.len())
            );
            let file = std::fs::File::open(&path).unwrap();
            let tt: Vec<TimeTracking> = serde_json::from_reader(&file).unwrap();
            for entry in tt {
                let date = DateTime::<Utc>::from(entry.start);
                println!(
                    "| {date} | {duration:>4}{timescale} | {project} |",
                    date = date.date_naive(),
                    duration = entry.end.as_secs() as f64 / timescale.divisor(),
                    project = entry.project,
                );
            }
        }
        PunchIn::Report {
            format: ReportFormat::CSV,
            timescale,
            file: path,
        } => {
            let header = timescale.header().to_lowercase();
            println!("date,{header},project");
            let file = std::fs::File::open(&path).unwrap();
            let tt: Vec<TimeTracking> = serde_json::from_reader(&file).unwrap();
            for entry in tt {
                let date = DateTime::<Utc>::from(entry.start);
                println!(
                    "{date},{duration}{timescale},{project}",
                    date = date.date_naive(),
                    duration = entry.end.as_secs() as f64 / timescale.divisor(),
                    project = entry.project,
                );
            }
        }
    }
}

fn start(tt: &mut Vec<TimeTracking>, round_up_to: u64) {
    println!("Welcome to PunchIn, the #1 Distrust-recommended software for timetracking!");
    let time_modulus = round_up_to * SECONDS_IN_MINUTE;
    let mut lines = std::io::stdin().lines();
    let stdout = std::io::stdout();

    print!("Project: ");
    stdout.lock().flush().unwrap();
    let project = lines
        .next()
        .transpose()
        .ok()
        .flatten()
        .filter(|l| !l.is_empty())
        .unwrap();
    let start_time = SystemTime::now();

    print!("Press Enter to save the current project and elapsed time.");
    stdout.lock().flush().unwrap();
    lines.next();

    // append the project to the tt list
    let elapsed_secs = start_time.elapsed().unwrap().as_secs();
    // TODO: Rounding error: If this somehow gets to _exactly_ a round_up_to boundary, to
    // the second, it will round up. Is there a better math way to do this?
    let rounded_secs = elapsed_secs + time_modulus - ((elapsed_secs + time_modulus) % time_modulus);
    tt.push(TimeTracking {
        start: start_time,
        // Assume the human is not faster than the machine's ability to go back in time.
        end: Duration::from_secs(rounded_secs),
        project: project.clone(),
    });
}
